from pathlib import Path

from loguru import logger
from omegaconf import OmegaConf

from .utils import resolve_napm_path


class NapmConfig:
    def __init__(
        self,
        env_name=None,
        config_path=None,
    ):
        self.env_name = env_name
        self._config_path = config_path


    @property
    def config_path(
        self
    ):
        """
        Returns the path to the config file.
        """
        if not self._config_path:
            self._config_path = Path(resolve_napm_path(self.env_name)) / 'config.yaml'
        logger.debug(f'{self._config_path}')
        if not Path(self._config_path).exists():
            #Path(self._config_path).touch()
            logger.debug(f'Creating config file at {self._config_path}')
            cfg = OmegaConf.create({'packages': {}})
            OmegaConf.save(cfg, self._config_path)
        return self._config_path


    def load(
        self
    ):
        """
        Loads the config file and returns a config object.
        """
        return OmegaConf.load(self.config_path)


    def update_config(
        self,
        config,
    ):
        """
        Updates the config file with the given config object.
        """
        OmegaConf.save(config, self.config_path)


    def add_package(
        self, 
        package_name, 
        install_dir,
        add_install_dir_to_path=False,
    ):
        """
        Adds a package to the config file.
        """
        config = self.load()
        config.packages[package_name] = {
            'install_dir': install_dir, 
            'add_install_dir_to_path': add_install_dir_to_path,
            }
        #config.packages[package_name]['install_dir'] = install_dir
        self.update_config(config)



#from distutils.command import install
import importlib
from pathlib import Path
from loguru import logger
#import os
import sys


from .utils import gitclone, resolve_napm_path
from .config import NapmConfig

def make_install_dir(dirname, env_name=None) -> str:
    """
    Creates the directory into which the package will be "installed".
    """
    parent = resolve_napm_path(env_name=env_name)
    install_dir = Path(parent) / dirname
    if not install_dir.exists():
        install_dir.mkdir(parents=True)
    else:
        logger.warning(f'{install_dir} already exists')
    install_dir = str(install_dir)
    return install_dir


def pseudoinstall_git_repo(
    package_url, 
    install_dir=None, 
    package_name=None,
    add_install_dir_to_path=False,
    env_name=None,
    ):
    """
    Clones a git repo, adds the install dir to `sys.path` if necessary, and
    then attempts to import the package
    
    :param package_url: the url of the git repo to clone
    :param install_dir: The directory where the package will be installed. If not specified, a directory
    will be created in the current working directory
    :param package_name: the name of the package you want to install
    :param add_install_dir_to_path: If the package is not found in the install dir, add the install dir
    to sys.path, defaults to False (optional)
    """
    install_successful = False
    if not package_name:
        package_name = package_url.split('/')[-1]
    if not install_dir:
        install_dir = make_install_dir(package_name, env_name=env_name)
    
    gitclone(package_url, install_dir)

    # test install was successful
    try:
        importlib.import_module(package_name)
        install_successful = True
    except ImportError as e:
        logger.error(f'{package_name} failed to import from dapm install dir')
        add_install_dir_to_path = True

    if add_install_dir_to_path:
        sys.path.append(install_dir)
        logger.debug(f"Added {install_dir} to sys.path")

    # validate that the package was "installed"
    try:
        importlib.import_module(package_name)
        install_successful = True
    except ImportError as e:
        logger.error(f"Failed to import {package_name}, napm 'install' failed")
        raise e
    
    if install_successful:
        config = NapmConfig(env_name=env_name)
        config.add_package(package_name, install_dir, add_install_dir_to_path)



import os
from pathlib import Path
import subprocess
import sys
#import importlib
#import omegaconf

from loguru import logger


# should we use an SDK for this instead? Maybe https://gitpython.readthedocs.io/en/stable/tutorial.html ?
def gitclone(
    repo_url, 
    tgt_dir,
    ):
    """
    Alias for calling git clone.
    """
    #cmd = ['git', 'clone']
    #cmd += [url, tgt_dir]
    cmd = ['git', 'clone', repo_url, tgt_dir]
    res = subprocess.run(cmd, stdout=subprocess.PIPE).stdout.decode('utf-8')
    #logger.debug(res)


def gitupdate(
    install_dir,
):
    """
    Updates a git repo.
    """
    # https://git-scm.com/docs/git#Documentation/git.txt--Cltpathgt
    logger.debug(f'attempting to invoke git pull on {install_dir}')
    cmd = ['git','-C',install_dir, 'pull', 'origin']
    res = subprocess.run(cmd, stdout=subprocess.PIPE).stdout.decode('utf-8')


#install_dir = Path(__path__)
#local_path = Path.cwd() 

# to do: add a database (maybe just a text file) to track mappings from package name to install dir


#Path.expanduser()

# TO DO: check for an enviroment variable that can be used to override the install directory


def resolve_napm_path(env_name=None) -> str:
    """
    Returns the parent direectory into which napm will "install" "pacakages"
    """
    # TO DO: better internal jargon to replace scarequotes
    # Todo: after first resolve, persist this to a config file or something like that and load from there. 
    napm_path = os.environ.get('NAPM_PATH')
    if napm_path is None:
        cache_dir = Path.home() / '.cache' / 'napm' 
        if env_name:
            cache_dir = cache_dir / '_envs' / env_name
        napm_path = (cache_dir).resolve()
        napm_path.mkdir(parents=True, exist_ok=True)
    napm_path = str(napm_path)
    if napm_path not in sys.path:
        sys.path.append(napm_path)
    return napm_path



from loguru import logger

from .pseudo_install import resolve_napm_path, pseudoinstall_git_repo
from .utils import gitclone, gitupdate, resolve_napm_path
from .config import NapmConfig

def set_autoupdate_flag(
    package_name,
    env_name=None,
    value=True,
):
    """
    Sets the "automated_update" flag for the specified "package" in the specified "environment".
    """
    config = NapmConfig(env_name=env_name).load()
    if package_name in config.get('packages'):
        config['packages'][package_name]['automated_update'] = value
        config.update_config(config)
    else:
        raise ValueError("Package not found in config.")


def update_package(
    package,
    env_name=None,
    config_path=None,
    force=False,
):
    """
    Updates (git pull) the specified "package" in the specified "environment".
    """
    logger.debug(package)
    if force:
        raise NotImplementedError
    config = NapmConfig(env_name=env_name, config_path=config_path)
    config = config.load()
    if package in config['packages']:
        install_dir = config['packages'][package]['install_dir']
        gitupdate(install_dir)
    else:
        raise ValueError("Package not found in config.")


def effect_automated_updates(
    env_name=None,
):
    """
    If packages have been previously "installed" with napm and flagged for automated update, 
    this will "update" the "packages" by attempting a git pull.
    """
    config = NapmConfig(env_name=env_name).load()
    for package_name, package_meta in config.packages.items(): # not sure if I can iterate cfg like this (codex)
        if package_meta.get('automated_update'):
            update_package(package_name, env_name=env_name)


def update_environment(
    env_name=None,
    ):
    """
    Updates all "packages" that have been "installed" into the specified "environment".
    """
    config = NapmConfig(env_name=env_name).load()
    for package_name, package_meta in config.packages.items(): # not sure if I can iterate cfg like this (codex)
        update_package(package_name, env_name=env_name)

# TO do: add an "auto-update" optional setting 
# to the config entry and pseudoinstall_git_repo function

# to do




from pathlib import Path
import sys

from loguru import logger

from .pseudo_install import resolve_napm_path
from .config import NapmConfig
from .update import effect_automated_updates


def populate_pythonpaths(
    env_name=None,
):
    """
    If packages have been previously "installed" with napm, this will add the "install" directories
    to the python path.
    """
    napm_paths = []
    napm_path = Path(resolve_napm_path())
    napm_paths.append(napm_path)
    if env_name:
        env_path = napm_path / env_name
        napm_paths.append(env_path)

    effect_automated_updates(env_name=env_name)
    config = NapmConfig(env_name=env_name).load()
    for package_name, package_meta in config.packages.items():
        if package_meta.add_install_dir_to_path:
            package_path = package_meta['install_dir']
            napm_paths.append(package_path)

    for path in napm_paths:
        if isinstance(path, Path):
            path = str(path.resolve())
        if path not in sys.path:
            sys.path.append(path)


if __name__ == "__main__":
    populate_pythonpaths()
